/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "InternalDataBuffer.h"
#include <cstring>

namespace xma {

#define XMA_DATA_BUFFER "DataBuffer"

//////////////////////////////////////////////////////////////////////////////
PrivateDataBufferInfo::PrivateDataBufferInfo(IntSessionFactoryPtr sf, LogPtr log) : m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "PrivateDataBufferInfo::PrivateDataBufferInfo()");
}

PrivateDataBufferInfo::~PrivateDataBufferInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "PrivateDataBufferInfo::~PrivateDataBufferInfo()");
}

bool PrivateDataBufferInfo::IsValid() const {
  return (m_dummy == DATA_BUFFER_DUMMY_VALUE);
}

std::recursive_mutex& PrivateDataBufferInfo::GetMutex() {
  return m_mutex;
}

void PrivateDataBufferInfo::SetSelf(DataBufferPtr self) {
  m_self = self;
}

DataBufferPtr PrivateDataBufferInfo::GetSelf() const {
  return m_self;
}

void PrivateDataBufferInfo::SetCallback(std::function<void(uint8_t*)> free_callback) {
  m_free_callback = free_callback;
}

void PrivateDataBufferInfo::CallCallback(uint8_t* data) {
  if (m_free_callback) {
    m_free_callback(data);
  }
}

IntSessionFactoryPtr PrivateDataBufferInfo::GetSessionFactory() {
  return m_sf;
}

LogPtr PrivateDataBufferInfo::GetLog() {
  return m_log;
}

//////////////////////////////////////////////////////////////////////////////
InternalDataBuffer::InternalDataBuffer(IntSessionFactoryPtr sf, LogPtr log, size_t size, bool dummy, int64_t pts, int64_t poc) : m_sf(sf), m_log(log) {
  if (dummy) {
    log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::InternalDataBuffer(dummy)");
  } else {
    log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::InternalDataBuffer(size=%lu)", size);
  }
  auto buf = std::make_unique<XmaDataBuffer>(); // will throw and not leak (defer assignment)

  memset(buf.get(), 0, sizeof(XmaDataBuffer));

  if (dummy) {
    buf->data.buffer_type = NO_BUFFER;
    buf->data.buffer      = nullptr;
    buf->alloc_size       = 0;
    buf->data.host        = nullptr;
  } else {
    buf->data.buffer_type = XMA_HOST_BUFFER_TYPE;
    buf->data.buffer      = new uint8_t[size]; // this *will* throw on OOM
    memset(buf->data.buffer, 0, size);
    buf->data.host  = static_cast<uint8_t*>(buf->data.buffer);
    buf->alloc_size = size;
  }
  buf->data.is_clone          = false;
  buf->pts                    = pts;
  buf->poc                    = poc;
  auto info                   = std::make_unique<PrivateDataBufferInfo>(sf, log);
  m_buf                       = std::move(buf); // past last exception risk
  m_buf->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_data_buffers.push_back(m_buf.get());
  lock.unlock();
#endif
  log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::InternalDataBuffer() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalDataBuffer::InternalDataBuffer(
    IntSessionFactoryPtr sf, LogPtr log, uint8_t* data, size_t size, std::function<void(uint8_t*)> free_callback, int64_t pts, int64_t poc) :
    m_sf(sf),
    m_log(log) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::InternalDataBuffer(size=%lu)", size);
  m_buf = std::make_unique<XmaDataBuffer>();

  memset(m_buf.get(), 0, sizeof(XmaDataBuffer));

  m_buf->data.buffer_type = XMA_HOST_BUFFER_TYPE;
  m_buf->data.buffer      = data;
  m_buf->data.is_clone    = true;
  m_buf->alloc_size       = size;
  m_buf->data.host        = static_cast<uint8_t*>(m_buf->data.buffer);
  m_buf->pts              = pts;
  m_buf->poc              = poc;

  auto info = std::make_unique<PrivateDataBufferInfo>(sf, log);
  info->SetCallback(free_callback);
  m_buf->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_data_buffers.push_back(m_buf.get());
  lock.unlock();
#endif
  log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::InternalDataBuffer() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalDataBuffer::~InternalDataBuffer() {
  if (!m_buf) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::~InternalDataBuffer(size=%d)", m_buf->alloc_size);

  DecRef();

  if ((m_buf) && (m_buf->data.refcount > 0)) {
    m_buf.release();
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::~InternalDataBuffer() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
XmaDataBuffer* InternalDataBuffer::GetDataBuffer() {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetDataBuffer", XMA_DATA_BUFFER, m_buf);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::GetDataBuffer(size=%d)", m_buf->alloc_size);

  return m_buf.get();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::IncRef() {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::IncRef", XMA_DATA_BUFFER, m_buf);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::IncRef(size=%d)", m_buf->alloc_size);

  ++m_buf->data.refcount;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::IncRef() - Returning");
  return m_buf->data.refcount;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::DecRef() {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::DecRef", XMA_DATA_BUFFER, m_buf);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::DecRef(size=%d)", m_buf->alloc_size);

  --m_buf->data.refcount;
  if (m_buf->data.refcount > 0) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::DecRef() - Returning");
    return m_buf->data.refcount;
  }

  if ((!m_buf->data.is_clone) && (m_buf->data.buffer)) {
    delete[] static_cast<uint8_t*>(m_buf->data.buffer);
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_data_buffers.remove(m_buf.get());
  lock.unlock();
#endif

  if (m_buf->private_do_not_touch) {
    PrivateDataBufferInfo* info = static_cast<PrivateDataBufferInfo*>(m_buf->private_do_not_touch);
    info->CallCallback(static_cast<uint8_t*>(m_buf->data.buffer));
    guard.unlock();
    delete info;
    m_buf->private_do_not_touch = nullptr;
  }

  m_buf->data.buffer = nullptr;
  m_buf              = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::DecRef() - Returning");
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::GetRefCount() const {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetRefCount", XMA_DATA_BUFFER, m_buf);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::GetRefCount(size=%d)", m_buf->alloc_size);

  return m_buf->data.refcount;
}

///////////////////////////////////////////////////////////////////////////////
void InternalDataBuffer::SetSelf() {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::SetSelf", XMA_DATA_BUFFER, m_buf);

  PrivateDataBufferInfo* info = static_cast<PrivateDataBufferInfo*>(m_buf->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    info->SetSelf(std::dynamic_pointer_cast<InternalDataBuffer>(shared_from_this()));
  }
}

///////////////////////////////////////////////////////////////////////////////
uint8_t* InternalDataBuffer::GetBuffer() const {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetBuffer", XMA_DATA_BUFFER, m_buf);

  return static_cast<uint8_t*>(m_buf->data.buffer);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::GetSize() const {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetSize", XMA_DATA_BUFFER, m_buf);

  return m_buf->alloc_size;
}

///////////////////////////////////////////////////////////////////////////////
int64_t InternalDataBuffer::GetPts() const {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetPts", XMA_DATA_BUFFER, m_buf);

  return m_buf->pts;
}

///////////////////////////////////////////////////////////////////////////////
int64_t InternalDataBuffer::GetPoc() const {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::GetPoc", XMA_DATA_BUFFER, m_buf);

  return m_buf->poc;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::SetSize(int32_t size) {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::SetSize", XMA_DATA_BUFFER, m_buf);

  if (size > m_buf->alloc_size) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_DATA_BUFFER, "InternalDataBuffer::SetSize() New size is larger than allocated size!");
    return XMA_ERROR_INVALID;
  }

  m_buf->alloc_size = size;

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::SetPts(int64_t pts) {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::SetPts", XMA_DATA_BUFFER, m_buf);

  m_buf->pts = pts;

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDataBuffer::SetPoc(int64_t poc) {
  LOCK_DATA_BUFFER(m_log, "InternalDataBuffer::SetPoc", XMA_DATA_BUFFER, m_buf);

  m_buf->poc = poc;

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t CopyDataBuffer(LogPtr log, XmaDataBuffer* dst, DataBufferPtr src) {
  if (!dst || (dst->data.buffer_type != NO_BUFFER) || !src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_DATA_BUFFER, "CopyDataBuffer() Invalid parameter!");
  }

  LOCK_DATA_BUFFER(log, "CopyDataBuffer", XMA_DATA_BUFFER, std::dynamic_pointer_cast<InternalDataBuffer>(src)->GetDataBuffer());

  log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "CopyDataBuffer(size=%d)", src->GetSize());

  IntDataBufferPtr src_buf = std::dynamic_pointer_cast<InternalDataBuffer>(src);
  if (!src_buf) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_DATA_BUFFER, "CopyDataBuffer() source not valid!");
  }

  PrivateDataBufferInfo* dst_info = static_cast<PrivateDataBufferInfo*>(dst->private_do_not_touch);
  if ((!dst_info) || (!dst_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_DATA_BUFFER, "CopyDataBuffer() Invalid private info!");
  }

  memcpy(dst, src_buf->m_buf.get(), sizeof(XmaDataBuffer));

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(src_buf->m_sf->m_mutex);
  src_buf->m_sf->m_data_buffers.remove(src_buf->m_buf.get());
  lock.unlock();
#endif

  IntDataBufferPtr dst_buf = std::dynamic_pointer_cast<InternalDataBuffer>(dst_info->GetSelf());
  src_buf->m_buf           = std::move(dst_buf->m_buf);
  delete dst_info;

  PrivateDataBufferInfo* src_info = static_cast<PrivateDataBufferInfo*>(src_buf->m_buf->private_do_not_touch);
  if ((!src_info) || (!src_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_DATA_BUFFER, "CopyDataBuffer() Invalid private info!");
  }
  src_info->SetSelf(src_buf);

  log->LogMsg(XMA_DEBUG_LOG, XMA_DATA_BUFFER, "CopyDataBuffer() - Returning");
  return XMA_SUCCESS;
}

}; // Namespace xma
