/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include "InternalSessionFactory.h"

namespace xma {

#define XMA_SESSION_BASE "SessionBase"

//////////////////////////////////////////////////////////////////////////////
class SessionBase : virtual public Session {
public:
  explicit SessionBase(IntSessionFactoryPtr sf) : m_sf(sf) { m_log = m_sf->GetLog(); }

  virtual ~SessionBase() {
  }

  virtual int32_t SetLog(LogPtr log) override {
    const char* level;
    switch (log->GetLogLevel()) {
    case XMA_EMERGENCY_LOG:
      level = "emergency";
      break;
    case XMA_ALERT_LOG:
      level = "alert";
      break;
    case XMA_CRITICAL_LOG:
      level = "critical";
      break;
    case XMA_ERROR_LOG:
      level = "error";
      break;
    case XMA_WARNING_LOG:
      level = "warning";
      break;
    case XMA_NOTICE_LOG:
      level = "notice";
      break;
    case XMA_INFO_LOG:
      level = "info";
      break;
    case XMA_DEBUG_LOG:
      level = "debug";
      break;
    default:
      level = "unknown";
      break;
    }
    switch (log->GetLogType()) {
    case XMA_LOG_TYPE_NONE:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=none)", level);
      break;
    case XMA_LOG_TYPE_CONSOLE:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=console)", level);
      break;
    case XMA_LOG_TYPE_SYSLOG:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=syslog)", level);
      break;
    case XMA_LOG_TYPE_FILE:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=file, file='%s')", level, log->GetFileName());
      break;
    case XMA_LOG_TYPE_CALLBACK:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=callback)", level);
      break;
    default:
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=unknown)", level);
      break;
    }
    m_log = log;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog() - Returning");
    return XMA_SUCCESS;
  }

  typedef VpiRet (*set_param_func)(void*, const char*, char*);

  XmaParameter* GetParameter(XmaParameter* params, int num_params, const char* name) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter(%s)", name);
    int i;
    for (i = 0; i < num_params; ++i) {
      if (!strcmp(name, params[i].name)) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter() - Returning parameter");
        return &params[i];
      }
    }
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter() - Returning NULL");
    return nullptr;
  }

  int32_t SetParam(XmaParameter* params, uint32_t param_cnt, const char* param_name, const int32_t min_value, const int32_t max_value, const int32_t default_value) {
    XmaParameter* param;
    if ((param = GetParameter(params, param_cnt, param_name)) != nullptr) {
      int32_t param_value = *reinterpret_cast<int32_t*>(param->value);
      if (param_value < min_value || param_value > max_value) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::SetParam() Invalid value for %s! (range = %d - %d)", param_name, min_value, max_value);
      } else {
        return param_value;
      }
    } else {
      return default_value;
    }
  }

  void SetParams(void* handle, XmaParameter* params, uint32_t param_cnt, const char* param_names[][2], uint32_t param_name_cnt, set_param_func func, const char* log_name,
      const char* err_msg_unkwn_param, const char* err_msg_unkwn_type) {
    uint32_t i, j;
    for (i = 0; i < param_cnt; ++i) {
      XmaParameter* param = &params[i];
      for (j = 0; j < param_name_cnt; ++j) {
        if (strcmp(param_names[j][0], param->name) == 0) {
          char text[64];
          switch (param->type) {
          case XMA_STRING:
            if (func(handle, const_cast<char*>(param_names[j][1]), static_cast<char*>(param->value)) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, static_cast<char*>(param->value));
            }
            break;
          case XMA_INT32:
            sprintf(text, "%d", *static_cast<int32_t*>(param->value));
            if (func(handle, const_cast<char*>(param_names[j][1]), text) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, text);
            }
            break;
          case XMA_UINT32:
            sprintf(text, "%u", *static_cast<uint32_t*>(param->value));
            if (func(handle, const_cast<char*>(param_names[j][1]), text) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, text);
            }
            break;
          case XMA_INT64:
            sprintf(text, "%ld", *static_cast<int64_t*>(param->value));
            if (func(handle, const_cast<char*>(param_names[j][1]), text) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, text);
            }
            break;
          case XMA_UINT64:
            sprintf(text, "%lu", *static_cast<uint64_t*>(param->value));
            if (func(handle, const_cast<char*>(param_names[j][1]), text) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, text);
            }
            break;
          case XMA_DOUBLE:
            sprintf(text, "%f", *static_cast<double*>(param->value));
            if (func(handle, const_cast<char*>(param_names[j][1]), text) < 0) {
              THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_param, param->name, text);
            }
            break;
          default:
            THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, log_name, err_msg_unkwn_type, param->type);
          }
        }
      }
    }
  }

protected:
  IntSessionFactoryPtr m_sf;
  LogPtr               m_log;
};

}; // Namespace xma
